// Copyright (C) 2006, 2007, 2008, 2009 by Florent Lamiraux, Thomas
// Moulard, Mathieu Poirier, CNRS.
//
// This file is part of the hppWalkPlanner.
//
// hppWalkPlanner is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or (at your option) any later version.
//
// hppWalkPlanner is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with hppWalkPlanner.  If not, see
// <http://www.gnu.org/licenses/>.

#include <sstream>
#include <boost/foreach.hpp>

#include <sys/time.h>
#include <libgen.h>

#include <KineoWorks2/kwsJointFreeFlyer.h>
#include <KineoModel/kppSMLinearComponent.h>

#include <kwsPlus/directPath/directPathVector.h>
#include <hpp/model/device.hh>
#include <kwsIO/kwsioConfig.h>

#include "pattern-generator.hh"

#include <hpp/util/debug.hh>

static const double HRP2_TIME_STEP = 0.005;

using namespace PatternGeneratorJRL;

namespace hpp
{
  namespace walkPlanner
  {
    PatternGenerator::PatternGenerator
    (CjrlHumanoidDynamicRobot* inRobot,
     double inStepLength,
     double inStepSpace) :
      attRobot(inRobot),
      attStepLength (inStepLength),
      attStepSpacing (inStepSpace),
      attStringStream ()
    {
      // Store initial configuration of robot at initialization
      // The corresponding posture will be recoved upon call of method
      // reset().
      attInitialPosture = inRobot->currentConfiguration();
    }

    void
    PatternGenerator::reset ()
    {
      //Set initial Position - half sitting position
      MAL_VECTOR_DIM (initialPosition, double,
		      attRobot->numberDof()-6);

      for(unsigned i = 0; i < attRobot->numberDof()-6; i++)
	{
	  initialPosition(i) =  attInitialPosture[i+6];
	}
      attWalkPatGen->SetCurrentJointValues (initialPosition);
    }

    PatternGenerator::~PatternGenerator ()
    {
      delete attWalkPatGen;
    }

    void
    PatternGenerator::initWalkPatGen()
    {
      // Create Pattern Generator Interface
      attWalkPatGen = patternGeneratorInterfaceFactory(attRobot);

      const char lBuffer[12][256] =
	{":samplingperiod 0.005",
	 ":previewcontroltime 1.6",
	 ":comheight 0.8078",
	 ":omega 0.0",
	 ":stepheight 0.07",
	 ":singlesupporttime 0.78",
	 ":doublesupporttime 0.02",
	 ":armparameters 0.5",
	 ":LimitsFeasibility 0.0",
	 ":ZMPShiftParameters 0.015 0.015 0.015 0.015",
	 ":TimeDistributionParameters 2.0 3.7 1.7 3.0",
	 ":UpperBodyMotionParameters -0.1 -1.0 0.0"
	};

      for(unsigned i = 0; i < 9; ++i)
	{
	  std::istringstream strm (lBuffer[i]);
	  attWalkPatGen->ParseCmd (strm);
	}
    }

    void
    PatternGenerator::computeStreamFromStepSequence
    (const footStepSequence_t& inRelativeFootPrintVector)
    {
      std::stringstream buffer;

      buffer << ":stepseq " ;
      BOOST_FOREACH (const footStep_t& fs, inRelativeFootPrintVector.footSteps ())
	buffer <<"\t" << fs[0]
	       <<"\t" << fs[1]
	       <<"\t" << fs[2];

      std::istringstream ibuffer (buffer.str ()); //FIXME: feeling depressed...
      attWalkPatGen->ParseCmd (ibuffer);
    }

    ktStatus
    PatternGenerator::computeAllQFromAStepSequence
    (hpp::model::DeviceShPtr hppDevice,
     CkwsPathShPtr& outPathHumanoid,
     std::vector<CkwsConfigShPtr>& outDiscretizedPath,
     std::vector<double>& HRP2TimeRef,
     std::vector<std::vector<double> >& HRP2ZMPRef,
     std::vector<std::vector<double> >& HRP2WaistPosition,
     const footStepSequence_t& inRelativeFootPrintVector)
    {
      CkwsConfigShPtr startConfig = CkwsConfig::create (hppDevice);
      CkwsSMLinearShPtr steeringMethod = CkwsSMLinear::create ();
      bool firstIteration = true;

      footStep_t initialFootStep = inRelativeFootPrintVector.initialFootStep ();
      double xInit = initialFootStep[0];
      double yInit = initialFootStep[1];
      double zInit = 0.;
      double thetaInit = initialFootStep[2];

      // Set the walking mode:
      //     - 0 : Normal mode.
      //     - 1 : Waist height variation.
      //     - 3 : Kineo works mode.
      //     - 4 : Following the upper-body motion provided by another plugin.
      // Specify the walking mode: here the default one.
      std::istringstream strm2 (":walkmode 0");
      attWalkPatGen->ParseCmd (strm2);

      computeStreamFromStepSequence (inRelativeFootPrintVector);

      unsigned long int NbOfIt = 0;

      MAL_VECTOR_DIM (currentAcceleration,double, 46);
      MAL_VECTOR_DIM (currentVelocity,double, 46);
      MAL_VECTOR_DIM (currentConfiguration, double, 46);
      MAL_VECTOR_DIM (ZMPTarget, double, 3);

      std::vector< CkwsDirectPathConstShPtr > dpVector;

      // Main Loop of PatternGenerator Computation
      while (attWalkPatGen->RunOneStepOfTheControlLoop
	     (currentConfiguration,
	      currentVelocity,
	      currentAcceleration,
	      ZMPTarget))
	{
	  ++NbOfIt;

	  // WalkGenJrl defines motions in the reference frame of the
	  // initial configuration of the robot
	  // We need to express the waist position in world frame.

	  double xPosShifted = xInit
	    + currentConfiguration (0) * std::cos (thetaInit)
	    - currentConfiguration (1) * std::sin(thetaInit);

	  double yPosShifted = yInit
	    + currentConfiguration (0) * std::sin(thetaInit)
	    + currentConfiguration (1) * std::cos(thetaInit);

	  double zPosShifted = zInit + currentConfiguration(2);

	  double rollPosShifted = currentConfiguration(3);
	  double pitchPosShifted = currentConfiguration(4);
	  double yawPosShifted = thetaInit + currentConfiguration(5);

	  currentConfiguration(0) = xPosShifted;
	  currentConfiguration(1) = yPosShifted;
	  currentConfiguration(2) = zPosShifted;
	  currentConfiguration(3) = rollPosShifted;
	  currentConfiguration(4) = pitchPosShifted;
	  currentConfiguration(5) = yawPosShifted;

	// Write in Data Structures

	std::vector<double> currentZMPTarget (3);
	std::vector<double> currentWaistPosition (6);
	std::vector<double> currentOpenHrpConfig (40);

	double time = HRP2_TIME_STEP * (HRP2TimeRef.size () + 1);
	HRP2TimeRef.push_back (time);
	hppDout (info, "HRP2TimeRef size= "
		 << HRP2TimeRef.size() << iendl
		 << "time = " << time);

	currentZMPTarget[0] = ZMPTarget(0);
	currentZMPTarget[1] = ZMPTarget(1);
	currentZMPTarget[2] = ZMPTarget(2);

	HRP2ZMPRef.push_back (currentZMPTarget);
	hppDout (info, "HRP2ZMPRef  size= " << HRP2ZMPRef.size ());

	// Record position of waist along motion
	for (unsigned i = 0; i < 5; ++i)
	  currentWaistPosition[i] = currentConfiguration (i);

	HRP2WaistPosition.push_back (currentWaistPosition);

	hppDout (info, "HRP2WaistPosition size= " << HRP2WaistPosition.size ());

	// Be careful do not insert
	// currentConfiguration(0)->currentConfiguration(5) : because
	// they are position and orientation

	for(unsigned count = 6 ;
	    count < MAL_VECTOR_SIZE (currentConfiguration); ++count)
	  currentOpenHrpConfig[count - 6] = currentConfiguration (count);


	// Express configuration in KineoWorks system of coordinates
	CkwsConfigShPtr endConfig  = CkwsConfig::create (hppDevice);
	std::vector<double> robotDofValues (hppDevice->countDofs ());

	if (!hppDevice->jrlDynamicsToKwsDofValues
	    (currentConfiguration, robotDofValues))
	  {
	    hppDout (info, "hppDevice: " << *hppDevice);
	    throw std::string("Failed to convert config from jrlDynamics to KineoWorks.");
	  }

	endConfig->setDofValues (robotDofValues);
	outDiscretizedPath.push_back (endConfig) ;
	hppDout (info, "outDiscretizedPath size= " << outDiscretizedPath.size ());

	if (firstIteration)
	  {
	    /*
	      At the first iteration, the initial configuration is set to
	      - the end configuration if the current path is empty
	      - the end configuration of the current path otherwise
	    */
	    if (dpVector.size () == 0)
	      for (unsigned int z = 0; z < endConfig->size (); ++z)
		startConfig->dofValue(z, endConfig->dofValue(z));
	    else
	      startConfig = dpVector.back ()->configAtEnd ();

	    firstIteration = false;
	  }
	else
	  {
	  // note : Kws can't built a Direct Path with startConfig =
	  //endConfig creation of the direct path

	    CkwsDPLinearShPtr DPLinear =
	      CkwsDPLinear::create(*startConfig, *endConfig, steeringMethod);

	    // If the robot has not moved, no path will be created, in
	    // that case, skip the end of the loop.
	    if (!DPLinear)
	      continue;

	    // check if the path is collising once for 2 configuration
	    if(NbOfIt % 2 == 0)
	      {
		hppDevice->setCurrentConfig (*endConfig);
		if(hppDevice->collisionTest ())
		  {
		    hppDout (error, "(self-)collision detected at t="
			      << NbOfIt * 0.005);
#ifndef HPP_DEBUG
		    outPathHumanoid.reset ();
#endif //! HPP_DEBUG
		    return KD_ERROR;
		  }
	      }

	  if (!DPLinear)
	    {
	      hppDout (error, "failed to create direct path");
	      hppDout (info, "startConfig=(" << *startConfig << ")");
	      hppDout (info, "endConfig=(" << *endConfig << ")");
#ifndef HPP_DEBUG
	      outPathHumanoid.reset ();
#endif //! HPP_DEBUG
	      return KD_ERROR;
	  }

	  // Add the direct path to the path.
	  dpVector.push_back (DPLinear);

	  hppDout (info, "outPathHumanoid length = " << dpVector.size());

	  for (unsigned z = 0; z < endConfig->size (); ++z)
	    startConfig->dofValue (z, endConfig->dofValue (z));
	  }
	}

      hppDout (notice, "number of iterations: " << NbOfIt);

      CdirectPathVectorShPtr directPath =
	CdirectPathVector::create (*(dpVector.front ()->configAtStart ()),
				   *(dpVector.back ()->configAtEnd()),
				   CkppSMLinearComponent::create (),
				   dpVector);
      if (outPathHumanoid->appendDirectPath (directPath) != KD_OK) {
	hppDout (error, "failed to insert direct path");
	return KD_ERROR;
      }
      return KD_OK;
    }

  } // end of namespace walkPlanner.
} // end of namespace hpp.
